using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace ShadowMapping
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        Model walls;
        Model columns;

        Effect lightEffect;

        Vector3 cameraPos = new Vector3(0, 2, 4.8f);
        Vector3 lightPosition;
        Vector3 lightDirection;

        Matrix view, projection;

        Camera camera;

        KeyboardState lastKbState = Keyboard.GetState();
        MouseState lastMsState = Mouse.GetState();

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            lightEffect = Content.Load<Effect>("ShadowMapping");

            walls = Content.Load<Model>("BathHouse-Walls");
            columns = Content.Load<Model>("BathHouse-Cols");

            camera = new Camera(cameraPos, -Vector3.UnitZ);
            view = camera.View;
            float aspectRatio = (float)graphics.GraphicsDevice.Viewport.Width / (float)graphics.GraphicsDevice.Viewport.Height;
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(75), aspectRatio, 0.1f, 100);

            lightEffect.Parameters["View"].SetValue(view);
            lightEffect.Parameters["Projection"].SetValue(projection);

            lightEffect.Parameters["cameraPosition"].SetValue(cameraPos);
            lightEffect.Parameters["specularPower"].SetValue(50);
            lightEffect.Parameters["specularIntensity"].SetValue(0);

            lightPosition = new Vector3(0, 2.5f, 0);
            lightDirection = new Vector3(-1.0f, -0.5f, -0.1f);
            lightDirection.Normalize();
            //--
            lightEffect.Parameters["lightPosition"].SetValue(lightPosition);
            lightEffect.Parameters["lightDirection"].SetValue(lightDirection);
            lightEffect.Parameters["cosInnerCone"].SetValue((float)Math.Cos(MathHelper.ToRadians(45)));
            lightEffect.Parameters["cosOutterCone"].SetValue((float)Math.Cos(MathHelper.ToRadians(60)));
            SetLightColors(lightEffect);

            UpdateLight();

            //--
            ChangeEffect(walls, lightEffect);
            ChangeEffect(columns, lightEffect);
        }

        private void SetLightColors(Effect effect)
        {
            effect.Parameters["ambientLightColor"].SetValue(new Vector4(0.1f, 0.1f, 0.1f, 1));
            effect.Parameters["diffuseLightColor"].SetValue(new Vector4(0.5f, 0.5f, 0.5f, 1));
            effect.Parameters["specularLightColor"].SetValue(new Vector4(1.0f, 1.0f, 1.0f, 1));
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            KeyboardState kbState = Keyboard.GetState();
            MouseState msState = Mouse.GetState();

            // Allows the game to exit
            if (kbState.IsKeyDown(Keys.Escape))
                this.Exit();

            if (kbState.IsKeyDown(Keys.Right))
            {
                lightDirection = Vector3.TransformNormal(lightDirection, Matrix.CreateRotationY(-0.1f));
                UpdateLight();
            }
            else if (kbState.IsKeyDown(Keys.Left))
            {
                lightDirection = Vector3.TransformNormal(lightDirection, Matrix.CreateRotationY(0.1f));
                UpdateLight();
            }

            float delta = (float)gameTime.ElapsedGameTime.TotalSeconds;

            // Camera Handling
            UpdateCamera(ref kbState, ref msState, delta);

            lastKbState = kbState;
            lastMsState = msState;

            base.Update(gameTime);
        }

        private void UpdateLight()
        {
            lightEffect.Parameters["lightDirection"].SetValue(lightDirection);
        }

        private void UpdateCamera(ref KeyboardState kbdState, ref MouseState msState, float frameSeconds)
        {
            float f = 0, r = 0;
            if (kbdState.IsKeyDown(Keys.W))
                f = frameSeconds;
            else if (kbdState.IsKeyDown(Keys.S))
                f = -frameSeconds;
            if (kbdState.IsKeyDown(Keys.D))
                r = frameSeconds;
            else if (kbdState.IsKeyDown(Keys.A))
                r = -frameSeconds;

            float rotUpDown = 0;
            float rotRightLeft = 0;

            if (lastMsState.RightButton == ButtonState.Pressed || msState.RightButton == ButtonState.Pressed)
            {
                rotUpDown = lastMsState.Y - msState.Y;
                rotRightLeft = lastMsState.X - msState.X;
            }

            if (f != 0 || r != 0 || rotUpDown != 0 || rotRightLeft != 0)
            {
                camera.Move(f, r);
                camera.Rotate(rotUpDown, rotRightLeft);
                view = camera.View;
                cameraPos = camera.Position;
                lightEffect.Parameters["View"].SetValue(view);
                lightEffect.Parameters["cameraPosition"].SetValue(cameraPos);
            }
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            DrawModel(walls, Matrix.Identity);
            DrawModel(columns, Matrix.Identity);

            base.Draw(gameTime);
        }

        private void DrawModel(Model model, Matrix world)
        {
            Matrix[] transforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(transforms);

            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (Effect effect in mesh.Effects)
                {
                    effect.Parameters["World"].SetValue(transforms[mesh.ParentBone.Index] * world);
                }
                mesh.Draw();
            }
        }

        public void ChangeEffect(Model model, Effect effect)
        {
            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (ModelMeshPart part in mesh.MeshParts)
                {
                    part.Effect = effect;
                }
            }
        }
    }
}
